from concurrent.futures import ProcessPoolExecutor
from typing import Iterable, Sequence, Callable

import numpy as np
from numpy import ndarray

from mldatasets.core.dataset import Dataset


class DataLoader(Iterable):
    """
    Data loader. Combines a dataset and a sampler, and provides an iterable over
    the given dataset.

    Arguments:
        dataset (Dataset): dataset from which to load the data.
        batch_size (int, optional): how many samples per batch to load
            (default: ``1``).
        shuffle (bool, optional): set to ``True`` to have the data reshuffled
            at every epoch (default: ``False``).
        max_workers (int, optional): how many subprocesses to use for data
            loading. ``0`` means that the data will be loaded in the main process.
            (default: ``0``)
        collate_fn (callable, optional): merges a list of samples to form a
            mini-batch of Tensor(s).  Used when using batched loading from a
            map-style dataset.
        drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
            if the dataset size is not divisible by the batch size. If ``False`` and
            the size of dataset is not divisible by the batch size, then the last batch
            will be smaller. (default: ``False``)
    """

    def __init__(
        self,
        dataset: Dataset,
        batch_size: int = 1,
        shuffle: bool = False,
        drop_last: bool = False,
        max_workers: int = 0,
        collate_fn: Callable = None,
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self._shuffle = shuffle
        self.drop_last = drop_last
        self.max_workers = min(max_workers, dataset.max_workers)
        self.collate_fn = default_collate if collate_fn is None else collate_fn

        self._batch_idx = 0
        self._sample_indices = np.arange(len(dataset))
        if shuffle:
            self.shuffle()

    def shuffle(self):
        self._sample_indices = np.random.permutation(len(self.dataset))

    def __len__(self) -> int:
        return len(self.dataset) // self.batch_size + int(self.drop_last)

    def __iter__(self):
        return self

    def _get_batch(self):
        start = self._batch_idx * self.batch_size
        end = start + self.batch_size

        if self.max_workers > 0:
            pool = ProcessPoolExecutor(max_workers=self.max_workers)
            samples = tuple(
                pool.map(self.dataset.__getitem__, self._sample_indices[start:end])
            )
        else:
            samples = tuple(self.dataset[i] for i in self._sample_indices[start:end])

        return self.collate_fn(samples)

    def __next__(self):
        if self._batch_idx < len(self):
            out = self._get_batch()
            self._batch_idx += 1
            return out
        else:
            self._batch_idx = 0
            if self._shuffle:
                self.shuffle()
            raise StopIteration


def default_collate(batch):
    r"""Puts each data field into a tensor with outer dimension batch size"""

    elem = batch[0]
    if isinstance(elem, (ndarray, float, int)):
        return np.array(batch)
    elif isinstance(elem, dict):
        return {key: default_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, Sequence):
        return tuple(default_collate(samples) for samples in zip(*batch))
    else:
        raise TypeError(
            f"Data type {type(elem)} not supported by default collate function. "
            f"Supported types: [np.ndarray, float, int, dict]."
        )
