from abc import ABC
from typing import Iterable, Callable

import numpy as np


class Dataset(ABC):
    def __init__(
        self,
        root: str,
        transform: Callable = None,
        target_transform: Callable = None,
        **kwargs
    ):
        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.max_workers = 2 ** 20

    def __len__(self):
        """Computes the length of the dataset (total number of samples)."""

    def __getitem__(self, index: int):
        """Retrieves a single sample from the dataset.  Each sample may contain
        multiple entries (e.g. image and label pair).
        """


class Subset(Dataset):
    def __init__(self, dataset: Dataset, indices: Iterable[int]):
        self._superset = dataset
        self.indices = list(indices)
        self.max_workers = dataset.max_workers

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, index: int):
        return self._superset[self.indices[index]]


class Concatenate(Dataset):
    def __init__(self, datasets: Iterable[Dataset]):
        self._subsets = tuple(datasets)
        self._start_indices = ()
        self._len = -1
        self.max_workers = min(d.max_workers for d in datasets)

    def __len__(self):
        if self._len < 0:
            self._len = sum(len(subset) for subset in self._subsets)
        return self._len

    @property
    def start_indices(self):
        if len(self._start_indices) == 0:
            lengths = [len(subset) for subset in self._subsets]
            self._start_indices = np.cumsum(lengths) - len(self._subsets[0])
        return self._start_indices

    def __getitem__(self, index: int):
        subset_idx = np.nonzero(index >= self.start_indices)[0][-1]
        offset = index - self.start_indices[subset_idx]
        return self._subsets[subset_idx][offset]
