import os
from typing import Callable, Optional

from mldatasets.core.dataset import Dataset


class DatasetFolder(Dataset):
    """A generic dataset where the images are arranged in this way:

        root/dog/xxx.png
        root/dog/xxy.png
        root/dog/xxz.png

        root/cat/123.png
        root/cat/nsdf3.png
        root/cat/asd932_.png

    Args:
        root (string): Root directory path.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, ``transforms.RandomCrop``
        target_transform (callable, optional): A function/transform that takes in the
            target and transforms it.
        loader (callable, optional): A function to load an image given its path.
    """

    def __init__(
        self,
        root: str,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        loader: Callable = None
    ):
        super().__init__(root, transform=transform, target_transform=target_transform)
        assert os.path.isdir(root), "Invalid dataset root: '{root}'."
        assert loader is not None, "Must provide a 'loader' function."
        self.loader = loader

        folders = filter(lambda f: os.path.isdir(os.path.join(root, f)), root)
        folder_to_idx = {os.path.join(root, f): i for i, f in enumerate(folders)}
        self.data = [
            (os.path.join(f, file_name), i)
            for f, i in folder_to_idx.items()
            for file_name in os.listdir(f)
        ]

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, index: int):
        path, y = self.data[index]
        x = self.loader(path)

        if self.transform is not None:
            x = self.transform(x)
        if self.target_transform is not None:
            y = self.target_transform(y)

        return x, y
